package ch.hevs.kart;

import android.os.Handler;

import java.util.EnumSet;
import java.util.LinkedList;
import java.util.List;

import ch.hevs.kart.comm.CommunicationSetup;
import ch.hevs.kart.comm.CommunicationSetupListener;
import ch.hevs.kart.comm.discovery.DiscoveryEntity;
import ch.hevs.kart.comm.protocol.Protocol;
import ch.hevs.kart.comm.protocol.ProtocolListener;
import ch.hevs.kart.comm.protocol.ProtocolState;
import ch.hevs.utils.Limits;

/**
 * Kart model. Once connected, the kart model allows to manipulate the kart developed during the
 * first year's summer school remotely simply by called methods on the instance. You do not need to
 * care about the communication with the kart at all, everything including retransmissions and error
 * handling is done for you.
 *
 * @author Nils Chatton
 * @author Christopher Metrailler (mei@hevs.ch)
 * @author Michael Clausen (clm@hevs.ch)
 * @version 2.2 - 2016
 */
@SuppressWarnings("WeakerAccess")
public class Kart {
	/**
	 * Maximal kart drive speed.
	 */
	public static final int MAXIMAL_SPEED =  15;
	
	/**
	 * Minimal drive speed (maximal backwards).
	 */
	public static final int MINIMAL_SPEED = -15;
	
	// Battery ADC has this value when batteries are 100% charged.
	// 1V ~ 200 => Fully charged (>13V) ~ 26000, empty (<11V) ~ 22000.
	private static final float MAXIMAL_BATTERIES_LEVEL = 26000f;
	private static final float MINIMAL_BATTERIES_LEVEL = 22000f;

	private static final int CONTROL_REGISTER_SYNC_INTERVAL_MS = 100;
	
	/*** Construction *****************************************************************************/
	
	/**
	 * Constructs a new kart object.
	 * <br><br>
	 * You probably do not have to construct the Kart instance if your activity is derived from
	 * {@link ch.hevs.kart.ui.AbstractKartControlActivity}, you can directly use the protected field
	 * {@link ch.hevs.kart.ui.AbstractKartControlActivity#kart} which refers to a connected Kart
	 * once your activity becomes active.
	 */
	public Kart() {
		// Save setup and subscribe to setup changes.
		this.setup = new KartSetup(setupListener);

		// Copy the setup values to the corresponding registers.
		setControlRegister(KartControlRegister.DrivePwmPeriod, setup.drivePwmPeriod());
		setControlRegister(KartControlRegister.SteeringStepPeriod, setup.steeringStepPeriod());
		setControlRegister(KartControlRegister.SteeringEndSwitchAddress,
					setup.steeringEndSwitchAddress());
		setControlRegister(KartControlRegister.HardwareSettings,
					KartHardwareSettings.toInt(setup.hardwareSettings()));
		setControlRegister(KartControlRegister.UpdateInterval, setup.updateIntervalMs());
		
		// Initialize other registers.
		setControlRegister(KartControlRegister.DriveSpeed, 0);
		setControlRegister(KartControlRegister.SteeringPosition, setup.steeringCenterPosition());
		setControlRegister(KartControlRegister.LEDs, 0);
		
		// Make all registers dirty in order to transmit them ASAP.
		uncommittedControlRegisters.addAll(EnumSet.allOf(KartControlRegister.class));
		
		// Register the main kart status register listener in order to forward the events to the
		// simpler kart listeners.
		addStatusRegisterListener(kartListenerDispatcher);
	}

	
	/*** Kart setup *******************************************************************************/
	
	/**
	 * Returns a reference to the kart setup.
	 * <br><br>
	 * Note that when you change settings on the kart setup object, these are applied immediately
	 * to the kart, so there is no need to call a setter method for kart setup changes.
	 *
	 * @return	Kart setup object for this kart instance.
	 */
	public KartSetup setup() {
		return setup;
	}

	// The actual kart setup object.
	private final KartSetup setup;

	// Kart setup listener processing setup changes.
	@SuppressWarnings("FieldCanBeLocal")
    private final KartSetup.KartSetupListener setupListener = new KartSetup.KartSetupListener() {
		@Override
		public void parameterChanged(final KartSetup.Parameter parameter) {
			switch (parameter) {
				case DrivePwmPeriod:
					setControlRegister(KartControlRegister.DrivePwmPeriod,
									   setup.drivePwmPeriod());
					break;

				case SteeringStepPeriod:
					setControlRegister(KartControlRegister.SteeringStepPeriod,
								setup.steeringStepPeriod());
					break;

				case SteeringEndSwitchAddress:
					setControlRegister(KartControlRegister.SteeringEndSwitchAddress,
								setup.steeringEndSwitchAddress());
					break;

				case SteeringMaxPosition:
					setSteeringPositionNormalized(0);
					break;

				case HardwareSettings:
					setControlRegister(KartControlRegister.HardwareSettings,
								KartHardwareSettings.toInt(setup.hardwareSettings()));
					break;

				case UpdateInterval:
					setControlRegister(KartControlRegister.UpdateInterval,
									   setup.updateIntervalMs());
					break;
			}
		}
	};


	/*** Communication ****************************************************************************/
	
	/**
	 * Tries to establish a connection to the given kart entity.
	 * <br><br>
	 * The operation is asynchronous and returns control direct to the caller. The status of the
	 * connection attempt is returned via the kart listener interface, method
	 * {@link KartListener#connectionStatusChanged(Kart, boolean)}.
	 *
	 * @param entity	Kart communication entity to connect to, a such entity can be found using
	 * 					the method
	 * 					{@link CommunicationSetup#startDiscovery}.
	 */
	public void connect(final DiscoveryEntity entity) {
		// Call the communications package asynchronous connect method.
		CommunicationSetup.connect(entity, protocolListener, commSetupListener);
	}
	
	/**
	 * Disconnects from the kart.
	 */
	public void disconnect() {
		// Stop updating registers to the kart.
		if(commitTask != null)
			commitTask.stop();
		
		// Close the communication with the kart.
		if (protocol != null)
			protocol.dispose();
		
		// Set protocol to null.
		protocol = null;
	}
	
	// CommunicationSetupListener interface implementation.
	private final CommunicationSetupListener commSetupListener = new CommunicationSetupListener() {
		@Override
		public void communicationSetupSucceeded(final Protocol protocol) {
			// Connection setup is ok, so save the pointer to the protocol.
			Kart.this.protocol = protocol;
			
			uncommittedControlRegisters.addAll(EnumSet.allOf(KartControlRegister.class));
			
			// Start synchronizing "dirty" registers to the kart.
			// TODO: Theoretically this is not anymore needed, but we need to test it before
			// 		 actually removing the background task...
			commitTask = new CommitTask(true);
			commitTask.start();
			
			// Signal new connection status to the kart listeners.
			for (KartListener listener: kartListeners) {
				listener.connectionStatusChanged(Kart.this, true);
			}
		}
		
		@Override
		public void communicationSetupFailed(final String reason) {
			// Signal failed connection attempt to all listeners.
			for (KartListener listener: kartListeners) {
				listener.message(Kart.this, "Connection setup error: " + reason);
			}
			
			// Signal new connection status to the kart listeners.
			for (KartListener listener: kartListeners) {
				listener.connectionStatusChanged(Kart.this, false);
			}
		}
	};
	
	// Protocol implementation/instance used to communicate with the kart.
	private Protocol protocol;

	// ProtocolListener implementation.
	private final ProtocolListener protocolListener = new ProtocolListener() {
		@Override
		public void registerValueIndication(final Protocol protocol,
                                            final KartStatusRegister register, int value) {
  		// Mask unsigned registers to avoid wrong negative values.
  		switch (register) {
			case SteeringPosition:
				if (value == 0 && steeringResetInProgress) {
					setup().clearHwSetting(KartHardwareSettings.RestartSlaves);
					steeringResetInProgress = false;
				}
			case HallSensorCounter1:
  		    case HallSensorCounter2:
  		    case BatteryVoltageLevel:
  		    case DistanceSensor:
			case SequenceStatus:
			case ProximitySensor1:
  		    case ProximitySensor2:
  		    case ProximitySensor3:
  		    case ProximitySensor4:
  		    case AmbientLightSensor1:
  		    case AmbientLightSensor2:
  		    case AmbientLightSensor3:
  		    case AmbientLightSensor4:
  		        value &= 0xFFFF;
  		        break;

			case SteeringStatus:
				value &= 0x0003;
				break;

			case ExternalEndSwitchState:
				value &= 0x0001;
				break;
  			}

			// Notify all listeners about the change.
			for (KartStatusRegisterListener listener: statusRegistersListeners)
				listener.statusRegisterHasChanged(Kart.this, register, value);

			// Update register value in local registers storage.
			statusRegisters[register.getAddress()] = value;
		}

		@Override
		public void stateChangeIndication(final Protocol protocol, final ProtocolState state) {
			// Is the protocol online?
			boolean online = state == ProtocolState.Online;
			
			// Inform all kart listeners about the change.
			for (KartListener listener: kartListeners) {
				listener.connectionStatusChanged(Kart.this, online);
			}
			
			// If we are not online anymore, close the connection.
			if (!online) {
				disconnect();
			}
		}

		@Override
		public void errorIndication(final Protocol protocol, final String error) {
			// Inform all kart listeners about the error.
			for (KartListener listener: kartListeners) {
				listener.message(Kart.this, "Communication error: " + error);
			}
		}
	};
	
	// Triggers a write of the control registers.
	private void commitControlRegisters() {
		if(commitTask != null)
			commitTask.commitHandler.post(new CommitTask(false));
	}

	// Task that updates the "dirty" registers of the local kart representation to the real kart.
	private class CommitTask implements Runnable {
		CommitTask(boolean continuously) {
			this.continuously = continuously;
		}
			
		@Override
		public void run() {
			// If the protocol is ready, we commit the registers.
			if (protocol != null && protocol.isReady()) {
				
				// Create a local copy of the set of registers that have to be written to the kart.
				EnumSet<KartControlRegister> uncommitted = uncommittedControlRegisters;
				
				// For each of the registers to commit, write it to the kart.
				for (KartControlRegister register: uncommitted) {
					protocol.setRegisterValue(register, controlRegisters[register.getAddress()]);
					synchronized (uncommittedControlRegisters) {
						uncommittedControlRegisters.remove(register);
					}
				}
        	}

			// Reschedule if task is continuously.
			if (continuously)
				commitHandler.postDelayed(this, CONTROL_REGISTER_SYNC_INTERVAL_MS);
		}

		public void start() {
			// Schedule the task for execution ASAP.
			commitHandler.post(this);
		}
		
		public void stop() {
			// Remove the task.
			commitHandler.removeCallbacks(this);
		}

		// If true, the task is executed continuously, otherwise only once.
		private final boolean continuously;
		public final Handler commitHandler = new Handler();
	}
	
	// The task actually updating the kart's registers.
	private CommitTask commitTask;


	/*** Convenience methods to enable a simple access to the kart. *******************************/
	
	/**
	 * Sets the drive speed (throttle) to the given absolute value.
	 * <br><br>
	 * Valid values are [-15, 15], whereas negative values make the kart driving backward and
	 * positive values forward.
	 *
	 * @param speed	Absolute drive speed (throttle).
	 */
	public void setDriveSpeed(final int speed) {
		setControlRegister(KartControlRegister.DriveSpeed,
					Limits.limit(speed, MINIMAL_SPEED, MAXIMAL_SPEED));
	}

	/**
	 * Sets the drive speed (throttle) to the given normalized value.
	 * <br><br>
	 * Valid values are [-1.0, 1.0], whereas negative values make the kart driving backward and
	 * positive values forward.
	 *
	 * @param speed Normalized drive speed (throttle).
	 */
	public void setDriveSpeedNormalized(final float speed) {
		setDriveSpeed((int) (speed * 15));
	}
	
	/**
	 * Increases the drive speed (throttle) by the smallest possible amount.
	 */
	public void increaseDriveSpeed() {
		setDriveSpeed(getDriveSpeed() + 1);
	}
	
	/**
	 * Decreases the drive speed (throttle) by the smallest possible amount.
	 */
	public void decreaseDriveSpeed() {
		setDriveSpeed(getDriveSpeed() - 1);
	}
	
	/**
	 * Returns the actual absolute drive speed (throttle) in the range [-15, 15].
	 *
	 * @return	Actual absolute drive speed.
	 */
	public int getDriveSpeed() {
		return getControlRegister(KartControlRegister.DriveSpeed);
	}
	
	/**
	 * Returns the actual normalized drive speed (throttle) in the range [-1.0, 1.0].
	 *
	 * @return	Actual normalized drive speed.
	 */
	public float getDriveSpeedNormalized() {
		return getControlRegister(KartControlRegister.DriveSpeed) / 15f;
	}
	
	/**
	 * Sets the steering position on the kart to the given absolute value.
	 * <br><br>
	 * Note that this is the set point of the steering control system on the kart. It may take some
	 * time for the steering system to actually reach the given position. The position is limited
	 * to the valid range from 0 to the maximal steering position in the kart setup.
	 *
	 * @param position	Absolute steering position.
	 */
	public void setSteeringPosition(final int position) {
		setControlRegister(KartControlRegister.SteeringPosition,
					Limits.limit(position, 0, setup.steeringMaxPosition()));
	}
	
	/**
	 * Sets the steering position on the kart to the given normalized value.
	 * <br><br>
	 * Note that this is the set point of the steering control system on the kart. It may take some
	 * time for the steering system to actually reach the given position. The position is limited
	 * to [-1, 1] whereas -1 and 1 are the left, respective right end positions.
	 *
	 * @param position	Normalized steering position.
	 */
	public void setSteeringPositionNormalized(final float position) {
		setSteeringPosition((int) (position * setup.steeringCenterPosition() +
                setup.steeringCenterPosition()));
	}
	
	/**
	 * Returns the actual absolute steering control position.
	 * <br><br>
	 * This is not the steering position the kart actually has, it is the position that the kart is
	 * instructed to reach.
	 *
	 * @return	Absolute steering position set point.
	 */
	public int getSteeringControlPosition() {
		return getControlRegister(KartControlRegister.SteeringPosition) & 0xFFFF;
	}
	
	/**
	 * Returns the actual normalized steering control position.
	 * <br><br>
	 * This is not the steering position the kart actually has, it is the position that the kart
	 * is instructed to reach.
	 *
	 * @return	Normalized steering position set point [-1.0, 1.0].
	 */
	public float getSteeringControlPositionNormalized() {
		return (float)getSteeringControlPosition() / (float)setup.steeringCenterPosition() - 1f;
	}
	
	/**
	 * Turns the steering position by the given absolute amount of steps.
	 * <br><br>
	 * The position is limited to the left and right end positions. Positive and negative numbers
	 * are possible whereas negative numbers signal opposite direction.
	 *
	 * @param amount	Absolute number of steps to turn.
	 */
	public void turnSteeringBy(final int amount) {
		setSteeringPosition(getSteeringControlPosition() + amount);
	}
	
	/**
	 * Turns the steering position by the given normalized amount.
	 * <br><br>
	 * The position is limited to the left and right end positions. Positive and negative numbers
	 * are possible whereas negative numbers signal opposite direction.
	 *
	 * @param amount	Normalized amount to turn.
	 */
	public void turnSteeringNormalizedBy(final float amount) {
		setSteeringPositionNormalized(getSteeringControlPositionNormalized() + 2f * amount);
	}

	/**
	 * Returns the actual absolute steering position (not the set point).
	 * <br><br>
	 * This is the position the steering system actually has and does not necessarily has to match
	 * the set point as the system needs some time to reach the requested positions.
	 *
	 * @return Actual absolute steering position.
	 */
	public int getSteeringPosition() {
		return getStatusRegister(KartStatusRegister.SteeringPosition) & 0xFFFF;
	}
	
	/**
	 * Returns the actual normalized steering position (not the set point) in the range [-1.0, 1.0].
	 * <br><br>
	 * This is the position the steering system actually has and does not necessarily has to match
	 * the set point as the system needs some time to reach the requested positions.
	 *
	 * @return Actual normalized steering position from [-1.0, 1.0].
	 */
	public float getSteeringPositionNormalized() {
		return (float)getSteeringPosition() / (float)setup.steeringCenterPosition() - 1f;
	}

	/**
	 * Returns true if the kart has reached the target steering position, false otherwise.
	 *
	 * @return True if the kart has reached the target steering position, false otherwise.
	 */
	public boolean isSteeringPositionReached() {
		return (getStatusRegister(KartStatusRegister.SteeringStatus) & 0x0002) != 0;
	}
	
	/**
	 * Returns true if the steering end contact is closed, false otherwise.
	 *
	 * @return True if steering end contact is closed, false otherwise.
	 */
	public boolean isSteeringEndContactTriggered() {
		return (getStatusRegister(KartStatusRegister.SteeringStatus) & 0x0001) != 0;
	}

    /**
     * Resets the steering subsystem of the kart.
	 * <br><br>
	 * Note that the method exits immediately while the initialization of the steering wheel system
	 * might take a moment.
     */
    public void resetSteering() {
        steeringResetInProgress = true;
        setup().setHwSetting(KartHardwareSettings.RestartSlaves);
    }

    private boolean steeringResetInProgress = false;

    /**
	 * Returns the battery level from 0.0 (empty) to 1.0 (full).
	 *
	 * @return Battery level [0.0, 1.0].
	 */
	public float getBatteryLevel() {
		return Limits.limit(((getStatusRegister(
				KartStatusRegister.BatteryVoltageLevel) & 0xFFFF) - MINIMAL_BATTERIES_LEVEL) /
				(MAXIMAL_BATTERIES_LEVEL - MINIMAL_BATTERIES_LEVEL), 0f, 1f);
	}
	
	/**
	 * Sets the LED or digital output at the give index [0,3] to the given value (true=state).
	 *
	 * @param index	LED index [0, 3].
	 * @param state	If true the LED is set state, if false the LED is set off.
	 */
	public void setLed(final int index, final boolean state) {
		if (index < 0 && index >= 4) return;
		int ledMask = getControlRegister(KartControlRegister.LEDs);
		if (state)
			ledMask |= 1 << index;
		else
			ledMask &= ~(1 << index);
		setControlRegister(KartControlRegister.LEDs, ledMask);
	}
	
	/**
	 * Gets the status if the LED/digital output at the given index.
	 *
	 * @param index	LED index [0, 3]
	 * @return		True if the LED is on, false if the LED is off.
	 */
	public boolean getLed(final int index) {
		if (index < 0 && index >= 4) return false;
		int ledMask = getControlRegister(KartControlRegister.LEDs);
		return (ledMask & (1 << index)) != 0;
	}
	
	/**
	 * Toggles the LED/digital output at the given index.
	 *
	 * @param index LED index [0, 3]
	 */
	public void toggleLed(final int index) {
		if (index < 0 && index >= 4) return;
		int ledMask = getControlRegister(KartControlRegister.LEDs);
		ledMask ^= 1 << index;
		setControlRegister(KartControlRegister.LEDs, ledMask);
	}
	
	
	/*** Control register API *********************************************************************/
	
	/**
	 * Sets the given control register to the given value.
	 * <br><br>
	 * Note that control registers can hold only 16 bit values, so not the whole range of Java's
	 * integer values is applicable.
	 *
	 * @see KartControlRegister
	 * @param register	The register to set.
	 * @param value		The value to set for the register.
	 */
	public void setControlRegister(final KartControlRegister register, final int value) {
		boolean changed = value != controlRegisters[register.getAddress()];
		controlRegisters[register.getAddress()] = value;
		if (changed) {
			synchronized (uncommittedControlRegisters) {
				uncommittedControlRegisters.add(register);
				commitControlRegisters();
			}
		}
	}
	
	/**
	 * Returns the actual value of the control register.
	 * <br><br>
	 * The value is casted signed to an int, so if you like to read the register value rather
	 * unsigned, use the {@link Kart#getControlRegisterUnsigned(KartControlRegister)} method.
	 *
	 * @param register	The control register to read.
	 * @return			The actual SIGNED register value.
	 */
	public int getControlRegister(final KartControlRegister register) {
		return controlRegisters[register.getAddress()];
	}

	/**
	 * Returns the actual value of the control register.
	 * <br><br>
	 * The value is casted unsigned to an int to avoid problems with big negative numbers.
	 *
	 * @param register	The control register to read.
	 * @return			The actual UNSIGNED register value.
	 */
	public int getControlRegisterUnsigned(final KartControlRegister register) {
		return controlRegisters[register.getAddress()] & 0xFFFF;
	}

	// Control register storage.
	private final int[] controlRegisters = new int[KartControlRegister.COUNT];
	
	// "Dirty" registers that have to be sent to the kart.
	private final EnumSet<KartControlRegister> uncommittedControlRegisters =
			EnumSet.allOf(KartControlRegister.class);


	/*** Status parameters ************************************************************************/
	
	/**
	 * Returns the actual value of the status register.
	 * <br><br>
	 * The value is casted signed to an int, so if you like to read the register value rather
	 * unsigned, prefer the {@link Kart#getStatusRegisterUnsigned(KartStatusRegister)} method.
	 *
	 * @param register	The status register to read.
	 * @return			The actual SIGNED register value.
	 */
	public int getStatusRegister(final KartStatusRegister register) {
		return statusRegisters[register.getAddress()];
	}
	
	/**
	 * Returns the actual value of the status register.
	 * <br><br>
	 * The value is casted unsigned to an int.
	 *
	 * @param register	The status register to read.
	 * @return			The actual UNSIGNED register value.
	 */
	public int getStatusRegisterUnsigned(final KartStatusRegister register) {
		return statusRegisters[register.getAddress()] & 0xFFFF;
	}

	// Status register storage.
	private final int[] statusRegisters = new int[KartStatusRegister.COUNT];
	
	
	/*** Status register listeners - low level interface ******************************************/
	
	/**
	 * Adds the given status listener to the kart.
	 * <br><br>
	 * The listener will be notified about the change of any status register on the kart after it
	 * has been registered using this method.
	 *
	 * @param listener	Reference to the listener to add.
	 */
	public void addStatusRegisterListener(final KartStatusRegisterListener listener) {
		statusRegistersListeners.add(listener);
	}
	
	/**
	 * Removes the given status register listener.
	 *
	 * @param listener Reference to the listener to remove.
	 */
	public void removeStatusRegisterListener(final KartStatusRegisterListener listener) {
		statusRegistersListeners.remove(listener);
	}
	
	// List of status register listeners.
	private final List<KartStatusRegisterListener> statusRegistersListeners =
			new LinkedList<KartStatusRegisterListener>();
	
	
	/*** Kart listeners - high level (limited) interface ******************************************/
	
	/**
	 * Adds the given kart listener to the kart.
	 * <br><br>
	 * The listener will be notified about the change of important values on the kart after it
	 * has been registered using this method.
	 *
	 * @param listener	Reference to the object implementing the kart listener interface.
	 */
	public void addKartListener(KartListener listener) {
		if (listener != null && !kartListeners.contains(listener))
			kartListeners.add(listener);
	}
	
	/**
	 * Removes the kart listener.
	 *
	 * @param listener Reference to the object implementing the kart listener interface to remove.
	 */
	public void removeKartListener(final KartListener listener) {
		if (listener != null) kartListeners.remove(listener);
	}
	
	// List of all kart listeners.
	private final List<KartListener> kartListeners = new LinkedList<KartListener>();
	
	// A status register implementation that notifies the Kart Listeners upon important events.
	@SuppressWarnings("FieldCanBeLocal")
    private final KartStatusRegisterListener kartListenerDispatcher = new KartStatusRegisterListener() {
		@Override
		public void statusRegisterHasChanged(final Kart kart, final KartStatusRegister register,
											 final int newValue) {
			switch (register) {
				case SteeringPosition:
					for (KartListener listener: kartListeners)
						listener.steeringPositionChanged(Kart.this, getSteeringPosition(),
								getSteeringPositionNormalized());
					break;
					
				case SteeringStatus:
					for (KartListener listener: kartListeners)
						listener.steeringPositionReachedChanged(Kart.this,
								isSteeringPositionReached());
					break;
					
				case BatteryVoltageLevel:
					for (KartListener listener: kartListeners)
						listener.batteryVoltageChanged(Kart.this, getBatteryLevel());
					break;

                case SequenceStatus:
					if (sequenceState == SequenceState.Busy && (newValue & 1)== 1) {
                        sequenceState = SequenceState.Running;
                    } else if (sequenceState == SequenceState.Running && (newValue & 1) == 0) {
                        protocol.setRegisterValue(KartControlRegister.SequenceControlRegister, 0);
                        sequenceState = SequenceState.Idle;
                        for (KartListener listener: kartListeners) {
                            listener.sequenceCompleted(Kart.this);
                        }
                    }

				default:
					break;
			}
		}
	};


	/*** Kart sequence API ************************************************************************/

	/**
	 * Returns true if a sequence is actually running (or downloaded) on the kart, false if no
	 * sequence is running.
	 * <br><br>
	 * Note that you can only start a sequence if there is not already a sequence running.
	 * <br><br>
	 * <b>IMPORTANT: This feature is experimental and might not be implemented on all Karts!</b>
	 * <br><br>
	 * Sequences can not be canceled at the moment.
	 *
	 * @return	True if sequence is running, false otherwise.
     */
	public boolean isSequenceRunning() {
		return sequenceState != SequenceState.Idle;
	}


	public void cancelRunningSequence() {
		protocol.setRegisterValue(KartControlRegister.SequenceControlRegister, 0x00);
	}

	/**
	 * Downloads the sequence to the kart and runs it autonomously on the kart.
	 * <br><br>
	 * Control to the kart is only returned after the sequence has been completed.
	 * <br><br>
	 * <b>IMPORTANT: This feature is experimental and might not be implemented on all Karts!</b>
	 * <br><br>
	 * Sequences can not be canceled at the moment.
	 *
	 * @param sequence	Sequence to download and start on the kart.
     */
	public void run(final KartSequence sequence) {
        if (sequenceState != SequenceState.Idle) return;
        sequenceState = SequenceState.Busy;
        int[] instructions = sequence.getKartVMInstructions();
		protocol.setRegisterValue(KartControlRegister.SequenceControlRegister, 0x01);
		for (int instruction: instructions) {
			protocol.setRegisterValue(KartControlRegister.SequenceFifoRegister, instruction);
		}
		protocol.setRegisterValue(KartControlRegister.SequenceFifoRegister,
				(new KartSequence.Instruction(KartSequence.Instruction.OpCode.END, (short)0)).getKartVMInstruction());
		protocol.setRegisterValue(KartControlRegister.SequenceControlRegister, 0x02);
    }

	// The actual sequence state.
	private enum SequenceState {Idle, Busy, Running}
	private SequenceState sequenceState = SequenceState.Idle;
}
