package ch.hevs.kart.ui;

import android.annotation.SuppressLint;
import android.app.Activity;
import android.content.Intent;
import android.content.pm.ResolveInfo;
import android.os.Bundle;
import android.view.ContextMenu;
import android.view.ContextMenu.ContextMenuInfo;
import android.view.Menu;
import android.view.MenuItem;
import android.view.View;
import android.widget.AdapterView;
import android.widget.AdapterView.OnItemClickListener;
import android.widget.ListView;
import android.widget.Toast;

import java.util.List;

import ch.hevs.kart.R;
import ch.hevs.kart.comm.CommunicationSetup;
import ch.hevs.kart.comm.discovery.Discovery;
import ch.hevs.kart.comm.discovery.DiscoveryEntity;
import ch.hevs.kart.comm.discovery.DiscoveryListener;

/**
 * Activity for scanning and displaying available Kart devices. This or a subclass should be the
 * start activity (the one added to the launcher) of the kart application.
 * 
 * @author Christopher Metrailler (mei@hevs.ch)
 * @author Michael Clausen (clm@hevs.ch)
 * @version 2.2 - 2016
 */
@SuppressLint("Registered")
public class EntityScanActivity extends Activity implements DiscoveryListener, OnItemClickListener {

	@Override
	public void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);

		// Get all available RCA Activities.
		Intent intent = new Intent(Intent.ACTION_RUN);
		intent.addCategory(AbstractKartControlActivity.INTENT_CATEGORY);
		rcActivities = getPackageManager().queryIntentActivities(intent, 0);

		// ListView Activity setup
		setContentView(R.layout.kart_discover_layout);
		foundDevices = (ListView)findViewById(R.id.list);
		foundDevices.setEmptyView(findViewById(R.id.empty));
		foundDevices.setOnItemClickListener(this);

		// Initializes list view adapter.
		adapter = new EntityListAdapter(this);
		foundDevices.setAdapter(adapter);
		registerForContextMenu(foundDevices);
	}

	@Override
	protected void onResume() {
		super.onResume();

        // Start automatically to scan for kart entities.
		scanDevice(true);
	}

	@Override
	protected void onPause() {
        // If scanning is actually in progress, stop it.
		if (scanningInProgress) {
			scanDevice(false);
		}

        // Clear list of found devices.
		adapter.clear();

		super.onPause();
	}

	@Override
	protected void onDestroy() {
		super.onDestroy();

        // If scanning is actually in progress, stop it.
		if (scanningInProgress) {
			scanDevice(false);
		}
	}

	@Override
	public boolean onCreateOptionsMenu(Menu menu) {
		getMenuInflater().inflate(R.menu.kart_discovery_menu, menu);
		if (!scanningInProgress) {
			menu.findItem(R.id.menu_stop).setVisible(false);
			menu.findItem(R.id.menu_scan).setVisible(true);
			menu.findItem(R.id.menu_refresh).setActionView(null);
		} else {
			menu.findItem(R.id.menu_stop).setVisible(true);
			menu.findItem(R.id.menu_scan).setVisible(false);
            MenuItem refreshItem = menu.findItem(R.id.menu_refresh);
            refreshItem.setActionView(R.layout.kart_discovery_actionbar);
            refreshItem.getActionView().setOnLongClickListener(new View.OnLongClickListener() {
                @Override
                public boolean onLongClick(View view) {
                    CommunicationSetup.enabledDummy();
                    if (scanningInProgress) {
                        scanDevice(false);
                        scanDevice(true);
                    }
                    return true;
                }
            });
		}

		return true;
	}

	@Override
	public boolean onOptionsItemSelected(MenuItem item) {
		switch (item.getItemId()) {
            case R.id.menu_scan:
                adapter.clear();
                scanDevice(true);
                break;
            case R.id.menu_stop:
                scanDevice(false);
                break;
		}
		return true;
	}

	private void scanDevice(final boolean enable) {
		if (enable) {
			scanningInProgress = true;
			CommunicationSetup.startDiscovery(this, this);
		} else {
			CommunicationSetup.stopDiscovery();
			scanningInProgress = false;
		}
		invalidateOptionsMenu();
	}

	@Override
	public void onItemClick(AdapterView<?> parent, View view, int position, long id) {
        // Select the entity that the user has clicked for later usage.
		if (!selectEntityAtIndex(position)) return;

        // If there are no registered activities to control a kart, show an error.
		if (rcActivities.isEmpty()) {
            Toast.makeText(this, R.string.error_no_activity, Toast.LENGTH_LONG).show();

        // If there is only one registered activity, just start that activity.
        } else if (rcActivities.size() == 1) {
            startControlActivity(rcActivities.get(0).activityInfo.name);

        // If there are multiple activities available, let the user choose one...
        } else {
            openContextMenu(foundDevices);
        }
	}

	@Override
	public void entityFoundIndication(Discovery discovery, final DiscoveryEntity entity) {
        // It is important that the adapter is updated on the UI thread.
		runOnUiThread(new Runnable() {
			@Override
			public void run() {
                // Add the entity to the UI.
				adapter.addEntity(entity);
			}
		});
	}

	@Override
	public void errorIndication(Discovery discovery, String error) {
        // Show any error from the discovery subsystem to the user.
		Toast.makeText(this, error, Toast.LENGTH_LONG).show();
	}

	@Override
	public void onCreateContextMenu(ContextMenu menu, View view, ContextMenuInfo menuInfo) {
		super.onCreateContextMenu(menu, view, menuInfo);

		// Display available RC activities
		for(int i = 0; i < rcActivities.size(); i++) {
			final ResolveInfo info = rcActivities.get(i);
			String label = info.activityInfo.loadLabel(getPackageManager()).toString();
			menu.add(0, i, i, label);
		}
	}

	@Override
	public boolean onContextItemSelected(MenuItem item) {
        // Start the control activity that the user has selected.
		startControlActivity(rcActivities.get(item.getItemId()).activityInfo.name);
		return true;
	}

	protected void startControlActivity(String className) {
        // Create the RC activity intent.
		final Intent intent = new Intent();
		intent.setClassName(this, className);

        // Pass the selected kart entity to the intent.
		intent.putExtra(AbstractKartControlActivity.EXTRAS_ENTITY, entity);

        // Change to the remote control activity.
		startActivity(intent);
	}

	protected boolean selectEntityAtIndex(final int index) {
        // Save the kart entity for later use in a field.
        entity = adapter.getEntity(index);

        // Return true if the saved entity is valid.
        return entity != null;
    }

	private ListView foundDevices;
	private EntityListAdapter adapter;
	private boolean scanningInProgress = false;
	private DiscoveryEntity entity;
	private List<ResolveInfo> rcActivities;
}