package ch.hevs.utils;

import android.os.Handler;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.Callable;

import ch.hevs.kart.Kart;

/**
 * Support for simple animations that run in the background (without blocking the Android App's
 * main thread).
 *
 * This can become quite handy if you just need to blink a LED or set an output to a given value for
 * a defined amount of time and then change it back.
 *
 * @author Michael Clausen (clm@hevs.ch)
 * @version 2.2 - 2016
 */
public class Animation {
    private interface Step {
        void process(Animation animation);
    }

    private static class ContinueStep implements Step {
        @Override
        public void process(final Animation animation) {
            animation.handler.post(new Runnable() {
                @Override
                public void run() {
                    animation.processNextStep();
                }
            });
        }
    }

    private static class WaitStep implements Step {
        private final int delay;

        public WaitStep(int delay) {
            this.delay = delay;
        }

        @Override
        public void process(final Animation animation) {
            animation.handler.postDelayed(new Runnable() {
                @Override
                public void run() {
                    animation.processNextStep();
                }
            }, delay);
        }
    }

    private static class ControlLedStep extends ContinueStep {
        private final int index;
        private final boolean state;

        public ControlLedStep(int index, boolean state) {
            this.index = index;
            this.state = state;
        }

        @Override
        public void process(Animation animation) {
            try {
                animation.kart.setLed(index, state);
            } catch (Exception e) {
                e.printStackTrace();
            }
            super.process(animation);
        }
    }

    private static class CustomStep extends ContinueStep {
        private final Runnable action;

        public CustomStep(final Runnable action) {
            this.action = action;
        }

        @Override
        public void process(Animation animation) {
            try {
                action.run();
            } catch (Exception e) {
                e.printStackTrace();
            }
            super.process(animation);
        }
    }

    private static class WaitUntilStep extends ContinueStep {
        private final Callable<Boolean> condition;
        private final int pollInterval;

        public WaitUntilStep(final Callable<Boolean> condition, int pollInterval) {
            this.condition = condition;
            this.pollInterval = pollInterval;
        }

        @Override
        public void process(final Animation animation) {
            try {
                boolean result = condition.call();
                if (result) {
                    super.process(animation);
                } else {
                    animation.handler.postDelayed(new Runnable() {
                        @Override
                        public void run() {
                            animation.repeatCurrentStepIn();
                        }
                    }, pollInterval);
                }
            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }

    private final Kart kart;
    private final Handler handler = new Handler();
    private final List<Step> steps = new ArrayList<Step>();
    private int currentStepIndex = 0;
    private boolean  isRunning= false;
    private boolean loop = false;

    private Animation(final Kart kart) {
        this.kart = kart;
    }

    private void addStep(final Step step) {
        steps.add(step);
    }

    private void repeatCurrentStepIn() {
        if (isRunning) {
            steps.get(currentStepIndex).process(this);
        }
    }

    private void processNextStep() {
        if (isRunning) {
            if (currentStepIndex < steps.size()) {
                steps.get(currentStepIndex++).process(this);
            } else {
                if (loop) {
                    currentStepIndex = 0;
                    processNextStep();
                } else {
                    isRunning = false;
                }
            }
        }
    }

    /**
     * Returns true if the animation is actually running and false if not.
     *
     * @return  True if running, false otherwise.
     */
    public boolean isRunnning() {
        return isRunning;
    }

    /**
     * Starts the animation and executes the animation exactly once. If the animation is already
     * running, the animation will be restarted from the beginning.
     */
    public void run() {
        if (isRunning) {
            handler.removeCallbacksAndMessages(null);
        }

        currentStepIndex = 0;
        isRunning = true;
        loop = false;

        processNextStep();
    }

    /**
     * Starts the animation and executes the animation in a loop infinitely. If the animation is
     * already running, the animation will be restarted from the beginning.
     */
    public void loop() {
        if (isRunning) {
            handler.removeCallbacksAndMessages(null);
        }

        currentStepIndex = 0;
        isRunning = true;
        loop = true;

        processNextStep();
    }

    /**
     * Cancels the animation. The animation is stopped and has to be restarted by using either the
     * run() or the loop() method.
     */
    public void cancel() {
        if (isRunning) {
            handler.removeCallbacksAndMessages(null);
        }

        currentStepIndex = 0;
        isRunning = false;
        loop = false;
    }

    /**
     * Builder in order to construct animations comfortable.
     */
    public static class Builder {
        private Animation animationToConstruct;

        private Builder(final Kart kart) {
            animationToConstruct = new Animation(kart);
        }

        /**
         * Turns the LED or digital output on the Kart at the given index on.
         *
         * @param index Index of the LED/Output to set to on (0..3).
         * @return      Reference to the builder in order to chain the method calls.
         */
        public Builder ledOn(int index) {
            animationToConstruct.addStep(new ControlLedStep(index, true));
            return this;
        }

        /**
         * Turns the LED or digital output on the Kart at the given index off.
         *
         * @param index Index of the LED/Output to set to off (0..3).
         * @return      Reference to the builder in order to chain the method calls.
         */
        public Builder ledOff(int index) {
            animationToConstruct.addStep(new ControlLedStep(index, false));
            return this;
        }

        /**
         * Waits the given amount of milliseconds before the next step of the animation will be
         * executed.
         *
         * @param milliseconds  Number of milliseconds to wait before executing the next step.
         * @return              Reference to the builder in order to chain the method calls.
         */
        public Builder wait(int milliseconds) {
            animationToConstruct.addStep(new WaitStep(milliseconds));
            return this;
        }

        /**
         * Performs a custom action that can be passed as Runnable implementation to the builder
         * method call.
         *
         * @param action    Runnable implementation that will be executed at the given animation
         *                  step.
         * @return          Reference to the builder in order to chain the method calls.
         */
        public Builder perform(final Runnable action) {
            animationToConstruct.addStep(new CustomStep(action));
            return this;
        }

        /**
         * Performs repeatedly the given condition until the condition returns true, then the
         * animation is continued. The default poll interval (the interval the condition is
         * called) is 100ms, thus a frequency of about 10 Hz.
         *
         * @param condition Callable implementation that will be executed at the given animation
         *                  step as long as it does not return true.
         * @return          Reference to the builder in order to chain the method calls.
         */
        public Builder until(final Callable<Boolean> condition) {
            animationToConstruct.addStep(new WaitUntilStep(condition, 100));
            return this;
        }

        /**
         * Performs repeatedly the given condition until the condition returns true, then the
         * animation is continued. The poll interval can be specified in milliseconds.
         *
         * @param condition     Callable implementation that will be executed at the given animation
         *                      step as long as it does not return true.
         * @param pollInterval  Interval at which's the condition is checked.
         * @return              Reference to the builder in order to chain the method calls.
         */
        public Builder until(final Callable<Boolean> condition, int pollInterval) {
            animationToConstruct.addStep(new WaitUntilStep(condition, pollInterval));
            return this;
        }

        /**
         * Finally builds the animation. This method will finalize the Animation instance building
         * process and returns a reference to the newly build animation.
         *
         * @return  Animation that was created using the builder.
         */
        public Animation build() {
            Animation animation = animationToConstruct;
            animationToConstruct = new Animation(animation.kart);
            return animation;
        }
    }

    /**
     * Returns a new Animation.Builder instance that can be used to build custom LED/Kart
     * animations.
     *
     * @param kart  Reference to the Kart instance that will be controlled by the animation.
     * @return      A reference to the newly created Animation builder instance.
     */
    public static Builder Builder(final Kart kart) {
        return new Builder(kart);
    }

    /**
     * The ExclusiveRunner class ensures that only exactly one animation is running at the same time
     * in its context. This means if the runner is instructed to run a new animation and there is
     * still an animation running in his context, the already runing animation will be stopped
     * before the new one is started.
     */
    public static class ExclusiveRunner {
        private Animation animation = null;

        /**
         * Starts the animation passed by animation and executes the animation exactly once. If
         * there is already an animation running in this instance, the animation will be canceled
         * first and then the new animation will be started.
         *
         * @param animation The animation to run.
         */
        public void run(final Animation animation) {
            if (this.animation != null) {
                this.animation.cancel();
            }
            this.animation = animation;
            this.animation.run();
        }

        /**
         * Starts the animation and executes the animation in a loop infinitely. If there is already
         * an animation running in this instance, the animation will be canceled first and then the
         * new passed animation will be looped until either the method cancel() will be called or a
         * new animation is set.
         *
         * @param animation The animation to loop.
         */
        public void loop(final Animation animation) {
            if (this.animation != null) {
                this.animation.cancel();
            }
            this.animation = animation;
            this.animation.loop();
        }

        /**
         * Cancels the actual animation that runs in the runner's context. The animation is stopped
         * and destroyed. In order to continue a new animation has to be restarted by using either
         * the run() or the loop() method.
         */
        public void cancel() {
            if (animation != null) {
                animation.cancel();
                animation = null;
            }
        }
    }
}