package ch.hevs.utils;

import android.os.Handler;
import android.os.Looper;

import ch.hevs.utils.Animation;

/**
 * Schedules one-shot or recurring tasks for execution.
 * <br><br>
 * You have to extend this abstract class and implement the method onTimeout().
 * <br><br>
 * The timer runs inside the Android main UI, so there will be no problems when using UI elements
 * from the timer's onTimeout() method. If the UI main thread is busy, timer execution may be
 * subject to delays.
 * <br><br>
 * One-shot timers are scheduled to run after a relative delay.
 * <br><br>
 * Periodically timers are scheduled with a fixed period and the very first time they are executed
 * immediately.
 * <br><br>
 * This class does not offer guarantees about the real-time nature of task scheduling. If you want
 * to animate LEDs or digital outputs you might consider to use the class
 * {@link Animation} which offers a very simple API to construct and run
 * animations.
 *
 * @author Michael Clausen (clm@hevs.ch)
 * @author Christopher Metrailler (mei@hevs.ch)
 * @version 2.2 - 2016
 */
public abstract class Timer {
    /**
     * This method will be called either once after the specified delay or periodically whenever
     * the timer expires.
     */
    public abstract void onTimeout();

    /**
     * Schedules the timer once to expire after the given delay.
     * <br>
     * Note that if the timer has already scheduled, this method will do nothing at
     *
     * @param delayMs The delay in milliseconds.
     */
    public void scheduleOnce(final int delayMs) {
        if (ready() && delayMs > 0) {
            interval = -1;
            handler.postDelayed(runnable, delayMs);
        }
    }

    /**
     * Schedules the timer continuously using the given interval. Note that the first time the timer
     * expires immediately.
     *
     * @param intervalMs Interval between the points in time the timer expires.
     */
    public void schedulePeriodically(final int intervalMs) {
        if (ready() && intervalMs > 0) {
            interval = intervalMs;
            handler.post(runnable);
        }
    }

    /**
     * Stops/cancels/pauses the pending timer. If the timer was scheduled once and did not already
     * expire, it will canceled. Periodically timers will be paused.
     * <br>
     * You can reschedule timers, but you need to wait until the ready() method returns true.
     */
    public void stop() {
        interval = 0;
        handler.removeCallbacks(runnable);
    }

    /**
     * If this method returns true, it is save to schedule the timer.
     *
     * @return True if the timer is ready to be rescheduled.
     */
    public boolean ready() {
        return interval == 0;
    }

    private int interval = 0;
    private final Handler handler = new Handler(Looper.getMainLooper());
    private final Runnable runnable = new Runnable() {
        @Override
        public void run() {
            if (interval != 0) {
                handler.postDelayed(runnable, interval);
            } else {
                interval = 0;
            }

            onTimeout();
        }
    };
}
